# make the code as Python 3 compatible as possible
from __future__ import print_function, division


import pypsa
import os
import pandas as pd
import numpy as np
import networkx as nx

os.path.join(os.path.dirname(pypsa.__file__), 'examples')

# In[2]:
basedir = os.path.join(os.path.dirname(pypsa.__file__), 'examples', 'opf-scigrid-de')
def read_scigrid(nodes_fn=os.path.join(basedir, 'scigrid-151109/vertices_de_power_151109.csvdata'),
                 links_fn=os.path.join(basedir, "scigrid-151109/links_de_power_151109.csvdata")):

    #note that some columns have 'quotes because of fields containing commas'
    vertices = pd.read_csv(nodes_fn, sep=",", quotechar="'", index_col=0)
    vertices["v_nom"] = 380.
    vertices.rename(columns={"lon":"x","lat":"y","name":"osm_name"}, inplace=True)
    vertices["x"] = pd.to_numeric(vertices["x"])
    vertices["y"] = pd.to_numeric(vertices["y"])
    print(vertices.columns)


    # In[4]:

    links = pd.read_csv(links_fn, sep=",", quotechar="'", index_col=0)
    links.rename(columns={"v_id_1":"bus0","v_id_2":"bus1","name":"osm_name"}, inplace=True)

    links["cables"].fillna(3,inplace=True)
    links["wires"].fillna(2,inplace=True)

    links["length"] = links["length_m"]/1000.

    default = dict(wires_typical=2.0, r=0.08, x=0.32, c=11.5, i=1.3)

    coeffs = {
            220000: dict(wires_typical=2.0, r=0.08, x=0.32, c=11.5, i=1.3),
            380000: dict(wires_typical=4.0, r=0.025, x=0.25, c=13.7, i=2.6)
        }

    links["r"] = [row["length"]*coeffs.get(row["voltage"],default)["r"]/(row["wires"]/coeffs.get(row["voltage"],default)["wires_typical"])/(row["cables"]/3.)  for i,row in links.iterrows()]

    links["x"] = [row["length"]*coeffs.get(row["voltage"],default)["x"]/(row["wires"]/coeffs.get(row["voltage"],default)["wires_typical"])/(row["cables"]/3.)  for i,row in links.iterrows()]

    links["s_nom"] = [3.**0.5*row["voltage"]/1000.*coeffs.get(row["voltage"],default)["i"]*(row["wires"]/coeffs.get(row["voltage"],default)["wires_typical"])*(row["cables"]/3.)  for i,row in links.iterrows()]


    # In[5]:

    print(links.columns)


    # In[6]:

    print(links["voltage"].value_counts(dropna=False))


    # In[7]:

    print(links[links["length_m"] <=0])


    # In[8]:

    print(links[(links["voltage"] != 220000) & (links["voltage"] != 380000)])


    # In[9]:

    print(links[pd.isnull(links.cables)])


    # In[10]:

    network = pypsa.Network()

    pypsa.io.import_components_from_dataframe(network,vertices,"Bus")

    pypsa.io.import_components_from_dataframe(network,links,"Line")


    # In[11]:

    network.build_graph()

    network.determine_network_topology()


    #remove small networks, which break the load flow
    for sn in network.sub_networks.itervalues():
        print(sn,len(sn.buses))
        if len(sn.buses) < 5:
            print(sn.branches,sn.buses)
            for bus in sn.buses.values():
                network.remove(bus)
            for branch in sn.branches.values():
                network.remove(branch)

    return network

network = read_scigrid()

network.lines_df.dtypes

def busmap_by_linemask(network, mask):
    lines = network.lines_df.obj.apply(lambda x: pd.Series(dict(bus0=x.bus0.name, bus1=x.bus1.name))) \
                            .assign(mask=mask) \
                            .set_index(['bus0', 'bus1'])
    G = nx.Graph()
    G.add_nodes_from(network.buses_df.index)
    G.add_edges_from(lines.index[lines['mask']])
    return pd.Series(dict((n, str(i))
                          for i, g in enumerate(nx.connected_components(G))
                          for n in g),
                     name='name')

def busmap_by_length(network, length):
    return busmap_by_linemask(network, network.lines_df.length < length)

busmap = busmap_by_length(network, 45.0)
busmap

def get_buses_linemap_and_lines(network, busmap):
    # compute new buses
    buses = network.buses_df \
        .groupby(busmap).agg(dict(v_nom=np.max, x=np.mean, y=np.mean)) \
        .assign(control='PQ') \
        .reindex_axis(network.buses_df.columns[:-1], axis=1)

    lines = pd.concat((network.lines_df, network.lines_df.obj.apply(lambda x: pd.Series(dict(bus0=x.bus0.name, bus1=x.bus1.name)))), axis=1)

    lines['bus0_s'] = lines.bus0.map(busmap)
    lines['bus1_s'] = lines.bus1.map(busmap)

    interlines = lines.ix[lines['bus0_s'] != lines['bus1_s']]

    def haversine(coords):
        lon, lat = np.deg2rad(np.asarray(coords)).T
        a = np.sin((lat[1]-lat[0])/2.)**2 + np.cos(lat[0]) * np.cos(lat[1]) * np.sin((lon[0] - lon[1])/2.)**2
        return 6371.000 * 2 * np.arctan2( np.sqrt(a), np.sqrt(1-a) )

    def aggregatelines(l):
        # l.name is a tuple of the groupby index (bus0_s, bus1_s)
        length_s = haversine(buses.loc[list(l.name),['x', 'y']])
        v_nom_s = consense(buses.loc[list(l.name),'v_nom'])
        voltage_factor = (np.asarray(network.buses_df.loc[l.bus0,'v_nom'])/v_nom_s)**2
        length_factor = length_s/l['length']
        # Do the aggregation by length rescaling everything

        #print("v_nom_s = {}\nv_nom = {}\nlength_s = {}\nlength = {}\n".format(v_nom_s, v_nom, length_s, l['length']))
        return pd.Series(dict(
                r=1./(voltage_factor/(length_factor * l['r'])).sum(),
                x=1./(voltage_factor/(length_factor * l['x'])).sum(),
                g=(voltage_factor * length_factor * l['g']).sum(),
                b=(voltage_factor * length_factor * l['b']).sum(),
                terrain_factor=l['terrain_factor'].mean(),
                s_nom=l['s_nom'].sum(),
                s_nom_min=l['s_nom_min'].sum(),
                s_nom_max=l['s_nom_max'].sum(),
                s_nom_extendable=l['s_nom_extendable'].any(),
                capital_cost=l['capital_cost'].sum(),
                length=length_s),
            index=list(l.columns[:-5]))

    lines = interlines.groupby(['bus0_s', 'bus1_s']).apply(aggregatelines)
    lines['name'] = [str(i+1) for i in range(len(lines))]
    linemap = interlines.join(lines['name'], on=['bus0_s', 'bus1_s'])['name']
    return (buses,
            linemap,
            lines.reset_index()
                 .rename(columns={'bus0_s': 'bus0', 'bus1_s': 'bus1'}, copy=False)
                 .set_index('name'))

buses, linemap, lines = get_buses_linemap_and_lines(network, busmap)

buses

linemap

lines

import matplotlib.pyplot as plt

from mpl_toolkits.basemap import Basemap

plt.plot([len(busmap_by_length(network, i).unique()) for i in np.arange(100)])

busmap = busmap_by_length(network, 40.0)
buses, linemap, lines = get_buses_linemap_and_lines(network, busmap)
lines = lines.set_index(['bus0', 'bus1'])

fig, ax = plt.subplots(figsize=(7,7))

x1 = 5
x2 = 15
y1 = 45
y2 = 58


bmap = Basemap(resolution='i',projection='merc',llcrnrlat=y1,urcrnrlat=y2,llcrnrlon=x1,urcrnrlon=x2)

bmap.drawcountries()
bmap.drawcoastlines()

for (f,t),row in lines.iterrows():

    lon = [buses.loc[f,"x"],buses.loc[t,"x"]]
    lat = [buses.loc[f,"y"],buses.loc[t,"y"]]

    x,y = bmap(lon,lat)

    color = "g"
    alpha = 0.7
    width =  1.2

    bmap.plot(x,y,color,alpha=alpha,linewidth=width)

x,y = bmap(buses["x"].values,buses["y"].values)

#highlight = known_points.keys()

#bmap.scatter(x,y,color=["r" if u in highlight else "b" for u in buses.index],s=[50 if u in highlight  else 2 for u in buses.index])

busmap.value_counts()

buses

lines

#%matplotlib inline

def build_network_from_buses_lines(buses, lines):
    network = pypsa.Network()

    pypsa.io.import_components_from_dataframe(network, buses, "Bus")
    pypsa.io.import_components_from_dataframe(network, lines, "Line")

    network.build_graph()
    network.determine_network_topology()
    
    return network

n2 = build_network_from_buses_lines(buses, lines)

n2.buses_df


